MCML <- function(X, y, u, BetaI, alphaI, csil=0.05, csiu=0.95, iterate=2, tol=1e-4, cutoff.adaptive=TRUE, gradient=TRUE, CoeffBound=9) {
# MCML estimate for given initial values BetaI and alphaI (set iterate=0 if no iteration is desired)
  p <- ncol(X)+1
  n <- length(y)
  BetaA <- BetaI
  alphaA <- alphaI
  muxA <- exp(drop(cbind(1,X)%*%BetaA))
  bound <- "no"
  BetaB <- rep(1000,p)
  alphaB <- 10000
  nit <- -1
  z <- rep(NA,n)
  while (max(abs(BetaA - BetaB)) > tol & nit < iterate & bound=="no") {
    nit <- nit+1
    BetaB <- BetaA
    muxB <- muxA 
    if (cutoff.adaptive) {
      ctf <- cutoffs.NB(y, u, alphaA, muxA, etal=(1-csil), etau=csiu)
      a <- ctf$a
      b <- ctf$b
      z <- ctf$z
    } else {
      a <- csil
      b <- csiu
    }
    Tax <- Tbx <- tax <- tbx <- rep(NA,n)
    for (j in 1:n) {
      Init <- Init.calc(muxA[j], alphaA, a, b)
      Tax[j] <- Init$ystar.a+2
      Tbx[j] <- Init$ystar.b
      tax[j] <- Init$g.a
      tbx[j] <- Init$g.b
    }
    par <- c(BetaA, sqrt(alphaA)) # reparametrize alpha
    if (gradient) {
      CML <- optim(par, fn=MCloglik, gr=GradMCML, method="BFGS", X=X, y=y, Tax=Tax, Tbx=Tbx, tax=tax, tbx=tbx)
    } else {
      CML <- optim(par, fn=MCloglik, gr=NULL, method="BFGS", X=X, y=y, Tax=Tax, Tbx=Tbx, tax=tax, tbx=tbx)
    }
    BetaA <- CML$par[1:p]
    alphaA <- CML$par[p+1]^2
    if (max(abs(BetaA)) > CoeffBound) {
      warning("bound in MCloglik", "\n")
      BetaA <- BetaB
      alphaA <- alphaB
      bound <- "yes"
    }
    muxA <- exp(drop(cbind(1,X)%*% BetaA))
  } # end while
  list(Beta=BetaA, alpha=alphaA, z=z, a=a, b=b, Tax=Tax, Tbx=Tbx, tax=tax, tbx=tbx, nit=nit)
}

#--- Scores
s1B <- function(y, mu, alpha) {
# mu score
  y/mu-(1+alpha*y)/(1+alpha*mu)
}

s2B <- function(y, mu, alpha) {
# alpha score
  Dg1 <- digamma(y+1/alpha)
  Dg2 <- digamma(1/alpha)
  DDD <- Dg1-Dg2 -log(alpha*mu+1)-alpha*(y-mu)/(alpha*mu+1) 
  (-1/alpha^2)*DDD
}

#--- Cutoff functions
Discr <- function(yo, Fo, cu) {
# minimum ratio between Fn and Fo (for argument yo > cu)
  n <- length(yo)
  Fn <- (0:(n-1))/n
  r <- Fn/Fo
  j <- (1:n)[yo > cu]
  if (length(j)>0) {
    j <- j[1]
    rj <- r[j:n]
  } else {
    j <- n+1
    rj <- 1
  }
  r.min <- min(rj,1)
  if (r.min < 1) {
    j.min <- (j:n)[rj==r.min]
  } else {
    j.min <- NULL
  }
  list(j=j, j.min=j.min, r.min=r.min)
}

Ptail <- function(y, u, alpha, mu) {
# randomized quantile residuals for NB(mu,alpha)
  p <- pnbinom(y, size=1/alpha, mu=mu)
  d <- dnbinom(y, size=1/alpha, mu=mu)
  p-u*d
}

cutoffs.NB <- function(y, u, alpha, mux, etal=0.95, etau=0.95) {
# cutoff values a and b for randomized quantile residuals in NB model
  z <- Ptail(y, u, alpha, mux)
  v <- sort(z)
  Fo <- punif(v)
  dsr <- Discr(v, Fo, etau)
  betu <- dsr$r.min
  tu <- as.numeric(quantile(v, probs=betu))
  tu <- max(tu, etau)
  v <- sort(-z + 1)
  Fo <- punif(v)
  dsr  <- Discr(v,Fo,etal)
  betl <- dsr$r.min
  tl <- as.numeric(quantile(v,probs=betl))
  tl <- max(tl,etal)
  tl <- -tl + 1
  betl <- 1-betl
  noise <- 1.0e-9
  list(z=z, a=tl+noise, b=tu-noise, betl=betl, betu=betu)
}

#--- Initial calculations 
Init.calc <- function(mu, alpha, a, b) {
  size <- 1/alpha
  ymax <- qnbinom(1-1e-6, size=size, mu=mu)
  ystar.a <- qnbinom(a, size=size, mu=mu)
  while (pnbinom(ystar.a, size=size, mu=mu) > a)
    ystar.a <- ystar.a-1
  ystar.b <- qnbinom(b, size=size, mu=mu)
  while (pnbinom(ystar.b, size=size, mu=mu) > b)
    ystar.b = ystar.b-1
  if (b==1)
    ystar.b <- ymax
  g.a <- (pnbinom(ystar.a+1, size=size, mu=mu) - a)/dnbinom(ystar.a+1, size=size, mu=mu)
  g.b <- (pnbinom(ystar.b+1, size=size, mu=mu) - b)/dnbinom(ystar.b+1, size=size, mu=mu)
  list(ystar.a=ystar.a, ystar.b=ystar.b, g.a=g.a, g.b=g.b)
}

QE <- function(mu, alpha, Ta, Tb, ta, tb) { 
# Probability of retained set
  size <- 1/alpha
  P.A <- pnbinom(Tb, size=size, mu=mu) - pnbinom(Ta-1, size=size, mu=mu) 
  p.Ta <- dnbinom(Ta-1, size=size, mu=mu) 
  p.Tb <- dnbinom(Tb+1, size=size, mu=mu)
  P.A + p.Ta*ta + p.Tb*(1-tb)
}

#--- Conditional loglikelihood
MCloglik <- function(par, X, y, Tax, Tbx, tax, tbx) {
# negative loglikelihood of MCML
# alpha is reparametrized on input
# X without column of ones 
  n <- length(y)
  p <- ncol(X)+1
  Beta <- par[1:p]
  alpha <- par[p+1]^2
  size <- 1/alpha
  mux <- exp(drop(cbind(1,X)%*%Beta))
#  print(summary(mux))
#  print(summary(Tbx))
#  print(summary(Tax))
#  print(size)
  P.Ax <- pnbinom(Tbx, size=size, mu=mux) - pnbinom(Tax-1, size=size, mu=mux)
  p.Tax <- dnbinom(Tax-1, size=size, mu=mux) 
  p.Tbx <- dnbinom(Tbx+1, size=size, mu=mux)
  Qex <- P.Ax + p.Tax*tax + p.Tbx*(1-tbx)
  ind.l <- which(y == Tax-1)          
  ind.u <- which(y == Tbx+1)
  ind.m <- which(y >= Tax & y <= Tbx)
  ind.H <- c(ind.l, ind.m, ind.u)
  pyx.l <- dnbinom(y[ind.l], size=size, mu=mux[ind.l], log=TRUE)
  pyx.u <- dnbinom(y[ind.u], size=size, mu=mux[ind.u], log=TRUE)
  pyx.m <- dnbinom(y[ind.m], size=size, mu=mux[ind.m], log=TRUE)
  L.l <- sum((pyx.l-log(Qex[ind.l]))*tax[ind.l])
  L.m <- sum(pyx.m-log(Qex[ind.m]))
  L.u <- sum((pyx.u-log(Qex[ind.u]))*(1-tbx[ind.u]))

  t.l <- sum(tax[ind.l]*log(tax[ind.l])) # this does not depend on alpha and Beta
  t.u <- sum((1-tbx[ind.u])*log(1-tbx[ind.u])) # this does not depend on alpha and Beta
  -(L.l + L.u + L.m + t.l + t.u)/n
}

#--- Gradient
GradMCML <- function(par, X, y, Tax, Tbx, tax, tbx) {
# Beta gradient of MCML estimate
# alpha is reparametrized on input
# X without column of ones 
  n <- length(y)
  p <- ncol(X)+1 
  Beta <- par[1:p]
  alpha <- par[p+1]^2
  size <- 1/alpha
  XI <- cbind(1, X)
  Etax <- drop(XI%*%Beta)
  Mux <- exp(Etax)
  HpBx <- exp(Etax)
  P.Ax <- pnbinom(Tbx, size=size, mu=Mux) - pnbinom(Tax-1, size=size, mu=Mux)
  p.Tax <- dnbinom(Tax-1, size=size, mu=Mux) 
  p.Tbx <- dnbinom(Tbx+1, size=size, mu=Mux)
  QEx <- P.Ax + p.Tax*tax + p.Tbx*(1-tbx)
  DmuQEx <- D.mu.QE(Mux, alpha, Tax,Tbx, tax,tbx, dmu=0.00001)
  DalQEx <- D.al.QE(Mux, alpha, Tax,Tbx, tax,tbx, dal=0.00001)
#
  ind.l <- which(y == Tax-1)          
  ind.u <- which(y == Tbx+1)
  ind.m <- which(y >= Tax & y <= Tbx)
  ind.H <- c(ind.l, ind.m, ind.u)
#
  s1B.l <- s1B(y[ind.l], Mux[ind.l], alpha)
  s1B.u <- s1B(y[ind.u], Mux[ind.u], alpha)
  s1B.m <- s1B(y[ind.m], Mux[ind.m], alpha)
  s2B.l <- s2B(y[ind.l], Mux[ind.l], alpha)
  s2B.u <- s2B(y[ind.u], Mux[ind.u], alpha)
  s2B.m <- s2B(y[ind.m], Mux[ind.m], alpha)
#
  QdmuD.l <- DmuQEx[ind.l]/QEx[ind.l]
  QdmuD.u <- DmuQEx[ind.u]/QEx[ind.u]
  QdmuD.m <- DmuQEx[ind.m]/QEx[ind.m]
  QdalD.l <- DalQEx[ind.l]/QEx[ind.l]
  QdalD.u <- DalQEx[ind.u]/QEx[ind.u]
  QdalD.m <- DalQEx[ind.m]/QEx[ind.m]
#
  z1B.l <- tax[ind.l]*(s1B.l - QdmuD.l)*HpBx[ind.l]
  z1B.u <- (1-tbx[ind.u])*(s1B.u - QdmuD.u)*HpBx[ind.u]
  z1B.m <- (s1B.m - QdmuD.m)*HpBx[ind.m]
#
  z2B.l <- tax[ind.l]*(s2B.l - QdalD.l)
  z2B.u <- (1-tbx[ind.u])*(s2B.u - QdalD.u)
  z2B.m <- (s2B.m - QdalD.m)
#
  zz1.H <- c(z1B.l, z1B.m, z1B.u)
  zz2.H <- c(z2B.l, z2B.m, z2B.u)
#
  XI.l <- XI[ind.l ,,drop=FALSE]
  XI.u <- XI[ind.u ,,drop=FALSE]
  XI.m <- XI[ind.m ,,drop=FALSE]
  XI.H <- rbind(XI.l, XI.m, XI.u)
#
  WWH <- diag(as.numeric(zz1.H))
  ss1 <- colSums(WWH%*%XI.H)
  ss2 <- sum(zz2.H)
  c(-ss1/n, -ss2/n)
}

#--- Information
InfoML <- function(Beta, alpha, X, gam=1e-7) {
# Information matrix of ML estimator
  p <- length(Beta)
  n <- nrow(X)
  A <- matrix(data=0, nrow=p, ncol=p)
  B <- matrix(data=0, nrow=p, ncol=1) 
  C <- 0
  XI <- cbind(1,X)
  size <- 1/alpha
  for (i in 1:n) {
    xi <- as.matrix(XI[i,])
    etaxi <- drop(Beta %*% xi)
    muxi <- hbxi <- exp(etaxi)
    k <- qnbinom(1-gam, size=size, mu=muxi)
    yik <- 0:k # k+1 values
    py <- dnbinom(yik, size=size, mu=muxi)
    py <- py/sum(py)
    s1 <- s1B(yik, muxi, alpha)
    s2 <- s2B(yik, muxi, alpha)
    E1 <- sum(s1^2*py)
    E12 <- sum(s1*s2*py)
    E2 <- sum(s2^2*py)
    A1 <- E1*hbxi^2 * xi%*%t(xi)
    B1 <- E12*hbxi * xi
    C1  <- E2
    A <- A + A1 
    B <- B + B1
    C <- C + C1
  }
# 
  I <- cbind(A, B)
  K <- cbind(t(B), C)
  Je <- rbind(I, K)/n
  Je
}

InfoMCML <- function(Beta, alpha, X, Tax, Tbx, tax, tbx, gam=1e-7) {
# Information matrix of modified CML estimator with fixed cutoffs
  p <- length(Beta)
  n <- length(Tax)
  A <- matrix(data=0, nrow=p, ncol=p)
  B <- matrix(data=0, nrow=p, ncol=1)
  C <- 0
  XI <- cbind(1,X)
  size <- 1/alpha
#
  for (i in 1:n) {
    xi  <- as.matrix(XI[i,])
    etaxi <- drop(Beta%*%xi)
    muxi <- exp(etaxi)
    Taxi <- Tax[i]
    Tbxi <- Tbx[i]
    taxi <- tax[i]
    tbxi <- tbx[i]
    QExi <- QE(muxi, alpha, Taxi, Tbxi, taxi, tbxi)
    DmuQExi <- D.mu.QE(muxi, alpha, Taxi, Tbxi, taxi, tbxi, dmu=0.00001)
    DalQExi <- D.al.QE(muxi, alpha, Taxi, Tbxi, taxi, tbxi, dal=0.00001)
    hbxi <- exp(etaxi)
    Dxi <- DmuQExi/QExi
    Fxi <- DalQExi/QExi
#
    k <- qnbinom(1-gam, size=size, mu=muxi)
    yik <- 0:k # k+1 values
#
    ind.l <- which(yik == Taxi-1)          
    ind.u <- which(yik == Tbxi+1)
    ind.m <- which(yik >= Taxi & yik <= Tbxi)
    ind.H <- c(ind.l, ind.m, ind.u)
#
    s1 <- s2 <- Py <- rep(0,(k+1))
    s1[ind.l] <- (s1B(yik[ind.l], muxi, alpha)-Dxi)*taxi
    s1[ind.u] <- (s1B(yik[ind.u], muxi, alpha)-Dxi)*(1-tbxi)
    s1[ind.m] <- (s1B(yik[ind.m], muxi, alpha)-Dxi)
#
    s2[ind.l] <- (s2B(yik[ind.l], muxi, alpha)-Fxi)*taxi
    s2[ind.u] <- (s2B(yik[ind.u], muxi, alpha)-Fxi)*(1-tbxi)
    s2[ind.m] <- (s2B(yik[ind.m], muxi, alpha)-Fxi)
#
    Py[ind.l] <- dnbinom(yik[ind.l], mu=muxi, size=size)*taxi/QExi
    Py[ind.m] <- dnbinom(yik[ind.m], mu=muxi, size=size)/QExi
    Py[ind.u] <- dnbinom(yik[ind.u], mu=muxi, size=size)*(1-tbxi)/QExi
#
###    py <- dnbinom(yik,size=size, mu=muxi); py <- py/sum(py)
###    py <- Py
#
    E1 <- sum(s1^2*Py)
    E12 <- sum(s1*s2*Py)
    E2 <- sum(s2^2*Py)
    A1 <- E1*hbxi^2 * xi%*%t(xi)
    B1 <- E12*hbxi * xi
    C1 <- E2
    A <- A + A1 
    B <- B + B1
    C <- C + C1
  }
  I <- cbind(A, B)
  K <- cbind(t(B), C)
  Je <- rbind(I, K)/n
  Je
}

#--- Derivatives
d.mu.fden <- function(y, mu, alpha) { 
# first derivative with respect of mu of the density
  size <- 1/alpha
  s1B(y, mu=mu, alpha)*dnbinom(y, mu=mu, size=size)
}

d.al.fden <- function(y, mu, alpha) {
# first derivative with respect of alpha of the density
  size <- 1/alpha
  s2B(y, mu=mu, alpha)*dnbinom(y, mu=mu, size=size)
}

d.mu.Fcum <- function(y, mu, alpha) {
# first derivative with respect of mu of the cdf
  k <- (0:y)
  size <- 1/alpha
  sum(s1B(k, mu=mu, alpha)*dnbinom(k, mu=mu, size=size))
}

d.al.Fcum <- function(y, mu, alpha) {
# first derivative with respect of alpha of the cdf
  k <- (0:y)
  size <- 1/alpha
  sum(s2B(k, mu=mu, alpha)*dnbinom(k, mu=mu, size=size))
}

D.mu.F <- function(y, mu, alpha, dmu=0.00001) {
# numerical first derivative with respect of mu of the cdf
  muu <- mu+dmu
  mud <- mu-dmu
  Fu <- pnbinom(y, mu=muu, size=1/alpha)
  Fd <- pnbinom(y, mu=mud, size=1/alpha)
  res <- (Fu-Fd)/(2*dmu)
  res
}

D.al.F <- function(y, mu, alpha, dal=0.00001) {
# numerical first derivative with respect of alpha of the cdf
  alu <- alpha+dal
  ald <- alpha-dal
  Fu <- pnbinom(y, mu=mu, size=1/alu)
  Fd <- pnbinom(y, mu=mu, size=1/ald)
  res <- (Fu-Fd)/(2*dal)
  res
}

d.mu.QE <- function(mu, alpha, Ta, Tb, ta, tb) {
  d.mu.Fcum(Tb, mu, alpha) - d.mu.Fcum(Ta-1, mu, alpha)+d.mu.fden(Ta-1, mu, alpha)*ta+d.mu.fden(Tb+1, mu, alpha)*(1-tb)
}

d.al.QE <- function(mu, alpha, Ta, Tb, ta, tb) {
  d.al.Fcum(Tb, mu, alpha)- d.al.Fcum(Ta-1, mu, alpha)+d.al.fden(Ta-1, mu, alpha)*ta + d.al.fden(Tb+1, mu, alpha)*(1-tb)
}

D.mu.QE <- function(mu, alpha, Ta, Tb, ta, tb, dmu=0.00001) { 
  muu <- mu+dmu
  mud <- mu-dmu
  Qeu <- QE(muu, alpha, Ta, Tb, ta, tb) 
  Qed <- QE(mud, alpha, Ta, Tb, ta, tb) 
  res <- (Qeu-Qed)/(2*dmu)
  res
}

D.al.QE <- function(mu, alpha, Ta, Tb, ta, tb, dal=0.00001) { 
  alu <- alpha+dal
  ald <- alpha-dal
  Qeu <- QE(mu, alu, Ta,Tb, ta, tb) 
  Qed <- QE(mu, ald, Ta,Tb, ta, tb) 
  res <- (Qeu-Qed)/(2*dal)
  res
}
